#!/usr/bin/env python

"""
Simplified AST nodes for easier type propagation and analysis.

Copyright (C) 2006 Paul Boddie <paul@boddie.org.uk>

This software is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License as
published by the Free Software Foundation; either version 2 of
the License, or (at your option) any later version.

This software is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public
License along with this library; see the file LICENCE.txt
If not, write to the Free Software Foundation, Inc.,
51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
"""

from compiler.visitor import ASTVisitor
import sys

# Unique name registration.

class Naming:

    "Maintain records of unique names for each simple name."

    def __init__(self):
        self.obj_to_name = {}
        self.names = {}

    def get(self, obj):
        return self.obj_to_name[obj]

    def set(self, obj, name):
        if self.obj_to_name.has_key(obj):
            return
        if not self.names.has_key(name):
            self.names[name] = 0
        n = self.names[name] + 1
        self.names[name] = n
        self.obj_to_name[obj] = "%s#%d" % (name, n)

naming = Naming()

def name(obj, name):
    naming.set(obj, name)
    return naming.get(obj)

# Elementary visitor support.

class Visitor(ASTVisitor):

    "A visitor base class."

    def __init__(self):
        ASTVisitor.__init__(self)

    def default(self, node, *args):
        raise ValueError, node.__class__

    def dispatch(self, node, *args):
        return ASTVisitor.dispatch(self, node, *args)

    def dispatches(self, nodes, *args):
        results = []
        for node in nodes:
            results.append(self.dispatch(node, *args))
        return results

# Simplified program nodes.

class Node:

    """
    A result node with common attributes:

    original    The original node from which this node was created.
    name        Any name involved (variable or attribute).
    index       Any index involved (temporary variable name).
    value       Any constant value.
    ref         Any reference to (for example) subprograms.
    nstype      Any indication of the namespace type involved in a name access.

    Expression-related attributes:

    expr        Any contributing expression.
    lvalue      Any target expression.
    test        Any test expression in a conditional instruction.

    Invocation and subprogram attributes:

    args        Any collection of argument nodes.
    params      Any collection of parameter nodes and defaults.

    Statement-grouping attributes:

    body        Any conditional code depending on the success of a test.
    else_       Any conditional code depending on the failure of a test.
    handler     Any exception handler code.
    finally_    Any code which will be executed regardless.
    code        Any unconditional code.
    choices     Any choices which may be included in the final program.
    """

    def __init__(self, original=None, **kw):

        """
        Initialise a program node with an optional link to an 'original' AST
        node.
        """

        self.original = original
        if self.original is not None:
            self.original._node = self
        for name, value in kw.items():
            setattr(self, name, value)

    def __repr__(self):
        if hasattr(self, "full_name"):
            s = "%s '%s'" % (self.__class__.__name__, self.full_name)
        elif hasattr(self, "name"):
            s = "%s '%s'" % (self.__class__.__name__, self.name)
        elif hasattr(self, "index"):
            s = "%s (%s)" % (self.__class__.__name__, self.index)
        elif hasattr(self, "value"):
            s = "%s %s" % (self.__class__.__name__, repr(self.value))
        elif hasattr(self, "ref"):
            s = "%s '%s'" % (self.__class__.__name__, name(self.ref, self.ref.name))
        else:
            s = "%s" % (self.__class__.__name__,)

        # Annotations.

        if hasattr(self, "types"):
            return "%s -> %s" % (s, self.types)
        else:
            return s

    def _pprint(self, indent, continuation, s, stream=None):
        stream = stream or sys.stdout
        if continuation:
            print >>stream, (" " * max(0, indent - len(continuation))) + continuation + s
        else:
            print >>stream, (" " * indent) + s

    def pprint(self, indent=0, continuation=None, stream=None):
        stream = stream or sys.stdout
        self._pprint(indent, continuation, repr(self), stream)

        # Subprogram-related details.

        if hasattr(self, "params"):
            for name, default in self.params:
                self._pprint(indent + 2, "( ", "%s -> %s" % (name, default), stream=stream)
            if hasattr(self, "star") and self.star:
                name, default = self.star
                self._pprint(indent + 2, "( ", "%s -> %s" % (name, default), stream=stream)
            if hasattr(self, "dstar") and self.dstar:
                name, default = self.dstar
                self._pprint(indent + 2, "( ", "%s -> %s" % (name, default), stream=stream)
        if getattr(self, "acquire_locals", 0):
            self._pprint(indent + 2, "( ", "acquiring locals", stream=stream)
        if getattr(self, "structure", 0):
            self._pprint(indent + 2, "( ", "structure '%s'" % self.structure.name, stream=stream)

        # Statement-related details.

        if hasattr(self, "test"):
            self.test.pprint(indent + 2, "? ", stream=stream)
        for attr in "code", "body", "else_", "handler", "finally_", "choices":
            if hasattr(self, attr) and getattr(self, attr):
                self._pprint(indent, "", "%s {" % attr, stream=stream)
                for node in getattr(self, attr):
                    node.pprint(indent + 2, stream=stream)
                self._pprint(indent, "", "}", stream=stream)

        # Expression-related details.

        if hasattr(self, "expr"):
            self.expr.pprint(indent + 2, "- ", stream=stream)
        if hasattr(self, "nodes"):
            for node in self.nodes:
                node.pprint(indent + 2, "- ", stream=stream)
        if hasattr(self, "lvalue"):
            self.lvalue.pprint(indent + 2, "->", stream=stream)
        if hasattr(self, "nstype"):
            self._pprint(indent + 2, "", self.nstype, stream=stream)
        if hasattr(self, "args"):
            for arg in self.args:
                arg.pprint(indent + 2, "( ", stream=stream)
            if hasattr(self, "star") and self.star:
                self.star.pprint(indent + 2, "( ", stream=stream)
            if hasattr(self, "dstar") and self.dstar:
                self.dstar.pprint(indent + 2, "( ", stream=stream)

        # Annotations.

        if hasattr(self, "accesses"):
            self._pprint(indent, "", "--------", stream=stream)
            for ref, attributes in self.accesses.items():
                self._pprint(indent + 2, "| ", "when %s: %s" % (ref, ", ".join([("%s via %s" % attr_acc) for attr_acc in attributes])), stream=stream)
            self._pprint(indent, "", "--------", stream=stream)
        if hasattr(self, "writes"):
            self._pprint(indent, "", "--------", stream=stream)
            for ref, attribute in self.writes.items():
                self._pprint(indent + 2, "| ", "when %s: %s" % (ref, attribute), stream=stream)
            self._pprint(indent, "", "--------", stream=stream)

class Pass(Node): "A placeholder node corresponding to pass."
class Return(Node): "Return an evaluated expression."
class Assign(Node): "A grouping node for assignment-related operations."
class Keyword(Node): "A grouping node for keyword arguments."
class Global(Node): "A global name designator."
class Import(Node): "A module import operation."
class LoadTemp(Node): "Load a previously-stored temporary value."
class LoadName(Node): "Load a named object."
class LoadAttr(Node): "Load an object attribute."
class LoadRef(Node): "Load a reference, typically a subprogram or a constant."
class LoadExc(Node): "Load a handled exception."
class StoreTemp(Node): "Store a temporary value."
class StoreName(Node): "Associate a name with an object."
class StoreAttr(Node): "Associate an object's attribute with a value."
class ReleaseTemp(Node): "Release a temporary value."
class Conditional(Node): "A conditional node consisting of a test and outcomes."
class Try(Node): "A try...except...else...finally grouping node."
class Raise(Node): "An exception raising node."
class Not(Node): "A negation of an expression."
class Choice(Node): "A special node which indicates a choice of expressions."

# Invocations involve some more work to process calculated attributes.

class Invoke(Node): "An invocation."

class InvokeFunction(Invoke):

    "A function or method invocation."

    def __init__(self, *args, **kw):
        Node.__init__(self, *args, **kw)
        if hasattr(self, "args"):
            self.set_args(self.args)

    def set_args(self, args):

        "Sort the 'args' into positional and keyword arguments."

        self.pos_args = []
        self.kw_args = []
        add_kw = 0
        for arg in args:
            if not add_kw:
                if not isinstance(arg, Keyword):
                    self.pos_args.append(arg)
                else:
                    add_kw = 1
            if add_kw:
                if isinstance(arg, Keyword):
                    self.kw_args.append(arg)
                else:
                    raise TypeError, "Positional argument appears after keyword arguments in '%s'." % self

class InvokeBlock(Invoke): "A block or loop invocation."

# Named nodes are those which can be referenced in some way.

class WithName:

    "Node naming."

    def __init__(self):
        self.full_name = name(self, self.name or "$untitled")

class Module(Node, WithName):

    "A Python module."

    def __init__(self, *args, **kw):
        Node.__init__(self, *args, **kw)
        WithName.__init__(self)

class Subprogram(Node, WithName):

    "A subprogram: functions, methods and loops."

    def __init__(self, *args, **kw):
        Node.__init__(self, *args, **kw)
        WithName.__init__(self)

# Special non-program nodes.

class Structure(Node): "A non-program node containing some kind of namespace."

class Class(Structure, WithName):

    "A Python class."

    def __init__(self, *args, **kw):
        Structure.__init__(self, *args, **kw)
        WithName.__init__(self)

class Instance(Structure):

    "An instance."

    def __repr__(self):
        return "Instance of type '%s'" % self.namespace.load("__class__")[0].type.full_name

class Constant(Instance):

    "A constant initialised with a type name for future processing."

    def __init__(self, *args, **kw):
        Instance.__init__(self, *args, **kw)
        self.typename = self.value.__class__.__name__

# vim: tabstop=4 expandtab shiftwidth=4
